/** 
 * @file Channel4Protocol.h
 * @brief Schnittstelle fuer die Klasse CChannel4Protocol.
 */
//////////////////////////////////////////////////////////////////////

#ifndef _CHANNEL4_PROTOCOL_H_
#define _CHANNEL4_PROTOCOL_H_

// Basisklasse einbinden.
#include "ChannelBase.h"
#include <time.h>

#include "../Common/crc.h"
#include "../Common/aes.h"
#include "TimeboyStation.h"

///////////////////////////////////////
// Defines zum Packet

// Makros fuer den Zugriff auf die Packetdaten
#define IV_PACKET_START(x) x[0]
#define IV_PACKET_DEVICEID(x) x[1]
#define IV_PACKET_DEVICEID_REMOTE(x) x[2]
#define IV_PACKET_SIZE_CONTROLBLOCK(x) x[3]
#define IV_PACKET_SIZE(x) (*(unsigned short*)&x[4])
#define IV_PACKET_FLAGS(x) x[6]
#define IV_PACKET_RESERVE(x) x[7]
#define IV_PACKET_CRC_HEAD(x) (*(unsigned short*)&x[8])

#define IV_PACKET_COMMAND(x) x[IV_PACKET_LENGTH_HEAD]
#define IV_PACKET_MODULE(x) (*(unsigned short*)&x[IV_PACKET_LENGTH_HEAD + IV_PACKET_LENGTH_CONTROLBLOCK + sizeof(unsigned char)])
#define IV_PACKET_FUNCTION(x) (*(unsigned short*)&x[IV_PACKET_LENGTH_HEAD + IV_PACKET_LENGTH_CONTROLBLOCK + sizeof(unsigned char) + sizeof(unsigned short)])

#define IV_PACKET_CRC(x) (*(unsigned short*)&x[IV_PACKET_LENGTH(x) - sizeof(unsigned short)])

// Grundangaben
#define IV_PACKET_LENGTH_MAX           0x300 // Maximalgroesse fuer ein Paketspeicher
#define IV_PACKET_LENGTH_HEAD          10    // Laenge des Packetkopfes
#define IV_PACKET_LENGTH_CONTROLBLOCK   1    // Laenge der Steuerinformation
#define IV_PACKET_LENGTH_FRAME         12    // Minimalframegroesse (Paketkopf + 16bit CRC)
#define IV_PACKET_LENGTH_DATA          (IV_PACKET_LENGTH_MAX - IV_PACKET_LENGTH_FRAME - IV_PACKET_LENGTH_CONTROLBLOCK)

// Paketsonderzeichen 
#define IV_PACKET_SIGN_START           0x7E  // ~ als Startzeichen der Pakete Byte 1
#define IV_PACKET_SIGN_ESCAPE          0x7F  // als Escapezeichen der Pakete

// Paketflags Byte 6
#define IV_PACKET_FLAG_DF4_KEY_T       0x01 // Verschluesselung Transfer-Key aktiv.
#define IV_PACKET_FLAG_DF4_KEY_OEM     0x02 // Verschluesselung OEM-Key aktiv.
#define IV_PACKET_FLAG_DF4_KEY_OEM2    0x04 // Verschluesselung OEM2-Key aktiv.
#define IV_PACKET_FLAG_FLASH_SERVICE   0x10 // Flash Service: Globaler Wear-Level ist >= 6				
#define IV_PACKET_FLAG_FLASH_WARNING   0x20	// Flash Warning: CRC Fehler im Ringspeicher oder Listenbereich. (Setup und Listen neu bertragen)		
#define IV_PACKET_FLAG_FLASH_LOCKED    0x40	// Flash Error: Keine freien Seiten mehr verfgbar, Gert befindet sich im Bedienstopp.
#define IV_PACKET_DF4_FIXVALUE         0xDAF0 // Pruefwert, ob Verschluesselung erfolgreich.

///////////////////
// Paketdaten

// Paketkommando (unteres Nibble)
#define IV_PACKET_VERSION_4            0x04 // Packetversion 4 seit FW 04.01.xx und Bootloader 04.xx
#define IV_PACKET_VERSION_5            0x05 // Packetversion 5 ist wie 4 jedoch mit Verschluesselung
// Paketkommando (oberes Nibble)
#define IV_PACKET_ANSWER               0x80 // Packet ist eine Antwort
#define IV_PACKET_REQUEST              0x40 // Packet ist eine Anfrage
#define IV_PACKET_ERROR                0x20 // Packet transportiert einen Fehler (Anfrage oder Antwort)


/////////////////////////////
// Werte fuer die Umgebung
 
// Standardwerte als Vorgabe
#define IV_DEFAULT_DEVICEID            32
#define IV_DEFAULT_DEVICEID_STANDALONE 0x00FE // Modbusadresse fuer Direktverbindung

#define IV_PENDING_TIMEOUT             25000 // Timeout fuer eine Aktive Antwort (ab 4.1.5.9 von 15 auf 25s)
#define IV_PENDING_FINGERPRINT_TIMEOUT 50000 // Timeout fuer eine Aktive Antwort (ab 4.2.1)
#define IV_PENDING_RESET_TIMEOUT       5000  // Timeout fuer eine Aktive Antwort bei SYSTEM.SYSTEMRESET (ab 4.2.1)
#define IV_PENDING_TRIALS(to)             (!(IV_PENDING_TIMEOUT % to) ? IV_PENDING_TIMEOUT / to : IV_PENDING_TIMEOUT / to + 1)
#define IV_PENDING_FINGERPRINT_TRIALS(to) (!(IV_PENDING_FINGERPRINT_TIMEOUT % to) ? IV_PENDING_FINGERPRINT_TIMEOUT / to : IV_PENDING_FINGERPRINT_TIMEOUT / to + 1)
#define IV_PENDING_RESET_TRIALS(to)       (!(IV_PENDING_RESET_TIMEOUT % to) ? IV_PENDING_RESET_TIMEOUT / to : IV_PENDING_RESET_TIMEOUT / to + 1)
#define IV_SKIPCOUNT                   1024 // Maximale Anzahl Zeichen die auserhalbt des Packetzusammenhangs zu lesen sind.

// Contexte fuer Execute-Routine
#define IV_CONTEXT_MASTER                    0x0000 // Ausfuehrung als Kommunikationsbefehl
#define IV_CONTEXT_SLAVE                     0x0001 // Pruefen auf anstehende Kommunikationsbefehle der Gegenstelle
#define IV_CONTEXT_TRANSMIT_ANSWER           0x0002 // Senden einer Antwort
#define IV_CONTEXT_TRANSMIT_ANSWER_ERROR     0x0004 // Senden einer Antwort als Fehler

// Maximale Laenge eines Fehlertextes
#define ERROR_TEXT_SIZE                      255

// Auswertung der Fehlerbestandteile
#define IV_RESULT_MODULE(res)                    ((res >> 8) & 0xFF)
#define IV_RESULT_FUNCTION(res)                  (res & 0xFF)


// Fuer die Verwendung des Kanals als ein Aktiver
enum ActiveState { 
	is_free = 0,      // Kanal ist nicht in Verwendung
	is_accepting, // Kanal nimmt eine Verbindung entgegen
	is_usable,    // Kanal hat die Verbindung angenommen und ermittelt Daten vom Gert.
	is_binding,   // Kanal durchluft die Kanalbindungsprfung.
	is_active     // Kanal ist Betriebsbereit.
};

/////////////////////////////
// Error's
#include "../Common/AEErrorDef.h"

//#define TRACEPACKET 1

///////////////////////////////////////
// Modul und Funktionsbasis

/** In diesem Module sind alle Kommunikationskontroll Funktionen enthalten.
 *
 * Es handelt sich hierbei um ein Modul zur Kontrolle der auszufuehrenden Funktionen
 * der abgeleiteten Module.
 *
 */
#define IV_MODULE_BASE                       ((short)0xFFFF)
// Undefinierte Modulfunktion
#define IV_FUNCTION_BASE                     ((short)0xFFFF)

/// Basisklasse fuer pure protokollbasierte Kommunikationsaufgaben.
class CChannel4Protocol : public CChannelBase
{
	friend class CAEChannel;
public:
	/// Konstruktor fuer Initialisierung der Klasse.
	CChannel4Protocol()  : CChannelBase(IV_DEFAULT_CHANNELID)
	{ 
		InitClass(); 
	}
	CChannel4Protocol(short channelID, short deviceID) : CChannelBase(channelID) 
	{ 
		InitClass(channelID, deviceID); 
	}
	virtual void InitClass(short channelID = IV_DEFAULT_CHANNELID, short deviceID = IV_DEFAULT_DEVICEID);
	/// Destruktor
	virtual ~CChannel4Protocol();

public:
	/// Oeffnen der Schnittstelle
	int Open(enumType type, char *string, int val, int to)
	{
		return CChannelBase::Open(type, string, val, to);
	}
	/// Schliessen der Schnittstelle.
	virtual int Close(void)
	{
		return CChannelBase::Close();
	}

// Speicher fuer Pakete/Daten
public:
	unsigned char m_transmit[IV_PACKET_LENGTH_MAX];
	unsigned char m_lastRequest[IV_PACKET_LENGTH_MAX];
	unsigned char m_receive[IV_PACKET_LENGTH_MAX];
	unsigned char m_data[IV_PACKET_LENGTH_DATA]; // 21 Byte Paketframe
	unsigned char m_tmp[IV_PACKET_LENGTH_MAX * 2];
	unsigned char *m_tmpPos;

	void Reset();

private:
	int m_packetStart;
	int m_skipCount;
	unsigned char m_packetNumber; // Paketzaehler
	unsigned char m_lastFlags;    // Halten der letzten empfangenen Flagvariable aus Packet. Fr Flash-Service Flags etc...
	double m_qs;

	clock_t m_clockTx, // Dauer des letzten Versendevorgangs
		m_clockRx,     // Dauer des letzten Empfangvorgangs
		m_clocksTx,    // Gesamtdauer der Versendevorgaenge
		m_clocksRx;    // Gesamtdauer der Empfangsvorgaenge
	int m_transmittedPackets;  // Anzahl gueltig versendeter Pakete
	int m_transmittedSizeData; // Anzahl der versendeten Nutzdaten-Bytes
	int m_receivedPackets;     // Anzahl gueltig empfangener Pakete
	int m_receivedPacketsReject;// Anzahl unbeantworteter Pakete
	int m_receivedSizeData;    // Anzahl der gelesenen Nutzdaten-Bytes

	short m_deviceID; // Die zu verwendende DeviceID der Klasse
	short m_deviceIDRemote; // Die zu verwendete DeviceID des Remote der Klasse
protected:
	// Membervariablen fuer Execute-Routine
	short m_lastModule;   // Zuletzt ausgefuehrtes Modul
	short m_lastFunction; // Zuletzt ausgefuehrte Funktion
	int m_result;         // Returnwert der Execute-Routine

// Routinen fuer DeviceID's
public:
	short GetDeviceID() 
	{ 
		return m_deviceID; 
	}
	short SetDeviceID(short deviceID)
	{ 
		short deviceIDPrev = m_deviceID; 
		m_deviceID = deviceID; 
		return deviceIDPrev; 
	}
	short GetDeviceIDRemote() 
	{ 
		return m_deviceIDRemote; 
	}
	short SetDeviceIDRemote(short deviceIDRemote)
	{ 
		short deviceIDPrev = m_deviceIDRemote; 
		m_deviceIDRemote = deviceIDRemote; 
		return deviceIDPrev; 
	}

// Routinen zur Ermittlung von sonstigen Memberwerten
public:
	inline unsigned char GetPacketNumber()
	{
		return m_packetNumber;
	}

// Standardroutinen
protected:
	// Zeiger auf CallBack-Funktionen
	int (*m_callBack1)(void);
	int (*m_callBack2)(unsigned int size, unsigned int count);
public:
	void SetCBFunction(int (*pFunction)(void));
	void SetCBFunctionWithParams(int (*pFunction)(unsigned int, unsigned int));

// Routinen fuer CRC 16 sowie 32bit
private:
	unsigned int m_crc32_table[256]; // Lookup table array
	void InitCRC32Table();
	unsigned int Reflect(unsigned int ref, char ch);
public:
	unsigned int GetCRC32(unsigned char *data, unsigned int size, unsigned int crc = 0xFFFFFFFF, unsigned char closure = TRUE);
	unsigned short GetCRC16(unsigned char *data, unsigned int size, unsigned short crc = 0xFFFF);

// Routinen fuer Verschluesselung der Paketdaten
private:
	unsigned char m_sbox[256], m_k[256], m_key[32];
protected:
	// Verschluesselungsroutinen
	void DF4LoadCryptKey(void);
	void DF4Init(int binaryKey);
	void DF4(unsigned char *data, int size);
	//int IsKeyOEM2() { return FALSE; }

protected:
	time_t
		m_timeReceivePacket,  // Zeitpunkt des letzten Packetempfangs
		m_timeTransmitPacket; // Zeitpunkt des letzten Packetversands
	// Packetdaten Byteweise schreiben und dabei Maskieren, komplettpacket senden.
	int  TransmitProtocol(unsigned char *buffer, int transmitTrials = IV_TRANSMIT_TRIALS);
	// Aufbereiten und senden eines Protokollpakets
	int TransmitPacket(short deviceIDRemote, unsigned char type, short module, short func, unsigned char *buf, int count, int transmitTrials = IV_TRANSMIT_TRIALS);

	// Packetdaten Byteweise empfangen und Demaskierung durchfuehren.
	int  ReceiveProtocol(unsigned char *buffer, int receiveTrials = IV_RECEIVE_TRIALS);
	int  ReceiveTimeboyStation(unsigned char *buffer, int receiveTrials = IV_RECEIVE_TRIALS);
	// Empfang und aufbereiten eines Protokollpakets
	int ReceivePacket(short deviceID, unsigned char &type, short &module, short &func, unsigned char *buf, int &count, int receiveTrials = IV_RECEIVE_TRIALS);

	//#################################
	// CallRoutinen fuer die Abarbeitung anstehender Kommunikationsbefehle
	virtual int Call(short module, short func, unsigned char *data, int &count);

public:
	// Verwaltungsklasse fuer Dockingstation im Channel
	CTimeboyStation m_timeboyStation;
	// Ermittlung ob es sich um einen Slot einer TimeboyStation handelt.
	int IsTimeboySlot(short deviceIDRemote)
	{
		return m_timeboyStation.IsSlot((unsigned char)deviceIDRemote);
	}
	// Beginn und Ende der Kommunikation ueber einen TimeboySlot
	virtual int SetTimeboySlotState(short deviceID, int state, int commandResult = IV_ERROR_SUCCESS)
	{
		return m_timeboyStation.SetTimeboySlotState(deviceID, state, commandResult);
	}
  void ResetTimeboyStation()
  {
    m_timeboyStation.ResetSlots();
  }
protected:
	unsigned int m_timeboyBaseAddress;

 
	// Konvertierung der Daten
public:
	void ULongToBinMSBFirst(unsigned char *Buf, unsigned int Val);
	void UIntToBinMSBFirst(unsigned char *Buf, unsigned int Val);
	unsigned int BinToULongMSBFirst(unsigned char *Dat);
	int BinToUIntMSBFirst(unsigned char *Buf);

public:
	// Ausfuehren eines Protokollbefehls
	virtual int Execute(short context, short module, short func, unsigned char *buf, int &count, int contextValue = -1);

private:
	volatile ActiveState m_activeState;

public:
	// Ist das Kanalobjekt in Verwendung als Aktive Verbindung.
	inline void SetActiveState(ActiveState as) 
	{ 
		m_activeState = as; 
	}
	inline ActiveState GetActiveState() 
	{ 
		return m_activeState; 
	}

// Fehlerbearbeitung
public:
	void ResetError();
	inline int GetLastResult() { return m_result; }
	// Fehlertexte fuer die Kommunikation/Module/Funktionen
	short m_errorCommunicationNum;
	char m_errorCommunication[ERROR_TEXT_SIZE];
	void GetErrorCommunication(short error, char *text);

	short m_errorModuleNum;
	char m_errorModule[ERROR_TEXT_SIZE];
	virtual void GetErrorModule(short module, short error, char *text);

	short m_errorFunctionNum;
	char m_errorFunction[ERROR_TEXT_SIZE];
	virtual void GetErrorFunction(short module, short func, short error, char *text);

	void TraceError(short deviceID, short module, short func);
};

#endif
