/** 
* @file TimeboyStation.h
* @brief Schnittstelle fuer die Klasse CTimeboyStation.
*/
//////////////////////////////////////////////////////////////////////

#ifndef _TIMEBOY_STATION_H_
#define _TIMEBOY_STATION_H_

// Basisklasse einbinden.
#include "ChannelBase.h"

//######################## TIMEBOY SLOT #######################################
// Defines Protokollbefehle
#define MB_READ		0x21		// bis zu 64 Byte lesen
#define MB_WRITE	0x22		// bis zu 64 Byte schreiben

// Defines fuer bestimmte Daten zu einem Slot
#define MB_SLOT_CHECK                  0x8000000C // Slot parameter fuer De/aktivierung der Timeboyerkennung
#define MB_SLOT_LED                    0x80001000 // Rote LED ein/aus
// Schreibe 1 Byte mit Wert 0 fuer: ein
// Schreibe 1 Byte mit Wert 255 fuer: aus
#define MB_SLOT_STATE                  0x8000E000 // Aktueller Steckzustand im Slot
#define MB_SLOT_VERSION                0x8000FFFF // Softwareversion der Slotfirmware

#define MB_SLOT_BUFFER_SIZE_SMALL      45  // Paketpuffergroesse bei FWVersion < 1.31
#define MB_SLOT_BUFFER_SIZE_LARGE      150 // Paketpuffergroesse bei FWVersion >= 1.31
#define MB_SLOT_BUFFER_DATA_SIZE_SMALL	(MB_SLOT_BUFFER_SIZE_SMALL - 13)
#define MB_SLOT_BUFFER_DATA_SIZE_LARGE	(MB_SLOT_BUFFER_SIZE_LARGE - 13)
//#############################################################################

// Datenstrucktur mit Daten zu einem Slot.
class CTimeboySlot  
{
public:
    CTimeboySlot()
    {
        Reset();
    }

public:
    void Reset()
    {
        // Member initialisieren
        m_version = 0;
        m_detectionState = 0;
        m_ledState = 0;
        m_ledStamp = 0;
        m_receiveStampSlot = 0;
        m_receiveStampTimeboy = 0;
    }
private:
    unsigned short m_version;          // Ermittelte Firmwareversion
public:
    void SetVersion(unsigned short version)	{ m_version = version; }
    unsigned short GetVersion() { return m_version; }
private:
    unsigned char m_detectionState;    // Ermittlung der Komperatorschaltung; 0 -> nicht ermittelt, 1 -> nicht verfuegbar, 2 -> verfuegbar
public:
    void SetDetectionState(unsigned char detectionState) { m_detectionState = detectionState; }
    unsigned char GetDetectionState() { return m_detectionState; }
private:
    unsigned char m_ledState;          // Zuletzt gesetzten LED-Zustand: 0x00 -> NA, 0x01 -> ON, 0xFF -> OFF
public:
    void SetLEDState(unsigned char ledState) { m_ledState = ledState; }
    unsigned char GetLEDState() { return m_ledState; }
private:
    clock_t m_ledStamp;                // clock() des letzten Zustand setzen.
public:
    void SetLEDStamp(clock_t ledStamp) { m_ledStamp = ledStamp; }
    clock_t GetLEDStamp() { return m_ledStamp; }
private:
    clock_t m_receiveStampSlot;        // clock() des zuletzt gueltig empfangenen Pakets.
public:
    void SetReceiveStampSlot(clock_t receiveStampSlot) { m_receiveStampSlot = receiveStampSlot; }
    clock_t GetReceiveStampSlot() { return m_receiveStampSlot; }
private:
    clock_t m_receiveStampTimeboy;     // clock() des zuletzt gueltig empfangene Antwort vom Timeboy
public:
    void SetReceiveStampTimeboy(clock_t receiveStampTimeboy) { m_receiveStampTimeboy = receiveStampTimeboy; }
    clock_t GetReceiveStampTimeboy() { return m_receiveStampTimeboy; }

};

// Automatische Jobs des Kommunikationsworkers
#define TSF_AUTO_LED         0x00000001 // Ausschalten der Slot-LED durch Worker, nach vorgegebenem Timeout
#define TSF_AUTO_CLOSE       0x00000002 // Schliessen der Schnittstelle, nach Abarbeitung aller Aufgaben.

// Klasse fuer die Verwaltung von Slots
class CTimeboyStation
{
public:
    CTimeboyStation()
    {
        // Member initialisiseren.
        m_channelBase = NULL;

        // Verbindungsthread
        m_communicationThread = NULL;
        m_threadState = end;
        m_threadCommand = none;

        // Anzahl der Suspendanforderungen halten
        m_suspendCounter = 0;

        // Flags fuer Steuerung der Idlezeit
        m_flags = 0;
    }
    virtual ~CTimeboyStation();

public:
    // Setzt die Basisklasse fuer die Kommunikationsschicht
    void SetChannelBase(CChannelBase& channelBase) { m_channelBase = &channelBase; }
    CChannelBase& GetChannelBase() 
    { 
        if (m_channelBase == NULL) throw; 
        return (CChannelBase&)*m_channelBase; 
    }

    inline int IsSlot(short deviceID)
    {
        return (GetSlotFirmwareVersion(deviceID) > 0);
    }

    unsigned char GetBufferSize(short deviceID)
    {
        // Bereich pruefen
        if (deviceID < 1 && deviceID > 255)
        {
            return 0;
        }

        // Je nach enthaltener Version
        switch(m_timeboySlot[deviceID].GetVersion()) 
        {
        case 0xFFFF: // Pruefung bereits durchgefuehrt und war nicht erfolgreich.
            break;

        case 0: // Ermittlung noch nicht durchgefuehrt
            return MB_SLOT_BUFFER_SIZE_SMALL;

        default:
            if (m_timeboySlot[deviceID].GetVersion() < 0x0131)
            {
                return MB_SLOT_BUFFER_SIZE_SMALL;
            }
            else if (m_timeboySlot[deviceID].GetVersion() >= 0x0131)
            {
                return MB_SLOT_BUFFER_SIZE_LARGE;
            }
        }

        return 0;
    }
    unsigned char GetDataBufferSize(short deviceID)
    {
        // Bereich pruefen
        if (deviceID < 1 && deviceID > 255)
        {
            return 0;
        }

        switch(m_timeboySlot[deviceID].GetVersion()) 
        {
        case 0xFFFF: // Pruefung bereits durchgefuehrt und war nicht erfolgreich.
            break;

        case 0: // Ermittlung noch nicht durchgefuehrt
            return MB_SLOT_BUFFER_DATA_SIZE_SMALL;

        default:
            if (m_timeboySlot[deviceID].GetVersion() < 0x0131)
            {
                return MB_SLOT_BUFFER_DATA_SIZE_SMALL;
            }
            else if (m_timeboySlot[deviceID].GetVersion() >= 0x0131)
            {
                return MB_SLOT_BUFFER_DATA_SIZE_LARGE;
            }
        }

        return 0;
    }

    // Main - Funktion fuer Befehlsausfuehrung in Workerthread
protected:
#ifdef _WIN32
    static unsigned long WINAPI CommunicationThread(LPVOID lpParam);
#else
    static void *CommunicationThread(void *lpParam);
#endif
#ifdef _WIN32
    HANDLE m_communicationThread;
    DWORD m_threadID;
#else
    pthread_t m_communicationThread;
#endif
protected:
    // Aktueller interner Zustand des Threads halten
    volatile enum threadState 
    {
        begin, // Bei Einstieg in Main-Routine
        end,   // Bei Ausstieg aus Main-Routine
        idle,  // Thread befindet sich im Leerlauf (laufende Pruefung auf anstehende Aufgaben)
        executeCommand, // Abarbeitung eines gesetzten Kommandos (threadCommand != none)
        dummy
    } m_threadState;
public:
    threadState GetThreadState() { return m_threadState; }
protected:
    // Auszufuehrendes Kommando
    volatile enum threadCommand
    {
        none,  // Aktuell kein abzuarbeitender Befehl
        exit,  // Beendigung des Threads erfordert
        write, // Schreiben von Bytes erfordert (Daten in m_commandParam)
        read   // Lesen von Bytes erfordert (Daten in m_commandParam)
    } m_threadCommand;
public:
    threadCommand GetThreadCommand() { return m_threadCommand; }
    int SetThreadCommand(threadCommand tc)
    {
        if (m_threadState == idle)
        {
            m_threadCommand = tc;
            return 1;
        }

        return 0;
    }
    volatile struct  
    {
        int res;
        short deviceID;
        unsigned int baseAddress;
        int offset;
        unsigned char *buf;
        int count;
        int trials;
    } m_commandParam;
    volatile int m_suspendCounter;
protected:
    int StartCommunicationThread();
    void StopCommunicationThread();

    // Gibt die FWVersion des Dockingslots an
    unsigned short GetSlotFirmwareVersion(short deviceID);
public:
    void ResetSlot(short deviceID);
    void ResetSlots();

public: // Routinen fuer die Kommunikation mit Docking
    int WriteBytes(short deviceID, unsigned int baseAddress, int offset, unsigned char *buf, int count, int transmitTrials = IV_TRANSMIT_TRIALS);
    int ReadBytes(short deviceID, unsigned int baseAddress, int offset, unsigned char *buf, int count, int receiveTrials = IV_RECEIVE_TRIALS);
protected:
    int ThreadWriteBytes(short deviceID, unsigned int baseAddress, int offset, unsigned char *buf, int count, int transmitTrials = IV_TRANSMIT_TRIALS);
    int ThreadReadBytes(short deviceID, unsigned int baseAddress, int offset, unsigned char *buf, int count, int receiveTrials = IV_RECEIVE_TRIALS);

public:
    int GetTimeboyModel(short deviceID, int &model);
    int SetTimeboySlotState(short deviceID, int state, int commandResult = IV_ERROR_SUCCESS);

protected:
    int ExecuteDocking(unsigned char *txBuf, unsigned char txCount, unsigned char *rxBuf, unsigned char *rxCount);
    int SetDockingLED(short deviceID, int state);

protected: // Routinen fuer die Flagverwaltung
    volatile unsigned int m_flags;
public:
    void AddFlag(unsigned int flag)
    {
        m_flags |= flag;
    }
    void DelFlag(unsigned int flag)
    {
        m_flags &= ~flag;
    }
    int IsFlag(unsigned int flag)
    {
        return ((m_flags & flag) != 0);
    }

protected:
    int m_autoLEDTimeout;
public:
    void SetLEDAutoControl(int enable, int timeout)
    {
        m_autoLEDTimeout = timeout;
        if (enable)
        {
            AddFlag(TSF_AUTO_LED);
        }
        else
        {
            DelFlag(TSF_AUTO_LED);
        }
    }

protected:
    CChannelBase *m_channelBase;
    CTimeboySlot m_timeboySlot[255];

    // Aktuell bearbeitete DeviceID halten
protected:
    short m_deviceIDRemote;
    inline void SetDeviceIDRemote(short deviceID)
    {
        m_deviceIDRemote = deviceID;
    }
    inline short GetDeviceIDRemote()
    {
        return m_deviceIDRemote;
    }

    // Helper zur Paketerstellung
protected:
    int BinToUIntMSBFirst(unsigned char *Buf)
    {
        return int(Buf[0] << 8) | int(Buf[1]);
    }
    unsigned int BinToULongMSBFirst(unsigned char *Dat)
    {
        unsigned int Val;

        Val  = Dat[0]; Val <<= 8;
        Val |= Dat[1]; Val <<= 8;
        Val |= Dat[2]; Val <<= 8;
        Val |= Dat[3];
        return Val;
    }

    unsigned short GetCRC16(unsigned char *data, unsigned int size, unsigned short crc = 0xFFFF);
};

#endif
